@@ -1,6 +1,7 @@
 import numpy as np
 from functools import reduce
 from scalib import _scalib_ext
+from scalib.config.threading import _get_threadpool
 
 
 class SASCAGraph:
@@ -10,7 +11,7 @@ class SASCAGraph:
     relationships between those variables and information about the values of
     the variables.
 
-    Variables have value in a binary finite field of size `nc`, and are viewed
+    Variables have values in a binary finite field of size `nc`, and are viewed
     as bit vectors. Those values are represented as integers in [0, nc).
     Variables can be qualified as `SINGLE` or `MULTI`, which relates
     to the multiple-execution feature of `SASCAGraph`: when performing a
@@ -48,7 +49,7 @@ class SASCAGraph:
             VAR MULTI x # Sbox input
             VAR MULTI y # Sbox output (whose leakage is targeted)
             PROPERTY x = k ^ p   # Key addition
-            PROPERTY x = sbox[y] # Sbox lookup
+            PROPERTY y = sbox[x] # Sbox lookup
             '''
         # n is the number of traces for our attack.
         graph = SASCAGraph(graph_desc,n)
@@ -164,14 +165,14 @@ class SASCAGraph:
         -------
         distribution : array_like, f64
             Distribution of `var`. If `var` is SINGLE, distribution has shape
-            `(1,nc)` else, has shape `(n,nc)`.
+            `(1,nc)`. Else, it has shape `(n,nc)`.
         """
         if not self.solved_:
             raise Exception("SASCAGraph not solved yet")
         return self.var_[var]["current"]
 
     def set_public(self, var, values):
-        r"""Marks a variable `var` has public with provided `values`.
+        r"""Marks a variable `var` as public with provided `values`.
 
         Parameters
         ----------
@@ -200,7 +201,7 @@ class SASCAGraph:
         self.publics_[var] = values
 
     def set_table(self, table, values):
-        r"""Defines a `table` content.
+        r"""Defines a `table`'s content.
 
         Parameters
         ----------
@@ -223,13 +224,15 @@ class SASCAGraph:
             )
         self.tables_[table] = values
 
-    def run_bp(self, it):
+    def run_bp(self, it, progress=False):
         r"""Runs belief propagation algorithm on the current state of the graph.
 
         Parameters
         ----------
         it : int
             Number of iterations of belief propagation.
+        progress: bool
+            Show a progress bar (default: False).
         """
         if self.solved_:
             raise Exception("Cannot run bp twice on a graph.")
@@ -241,6 +244,8 @@ class SASCAGraph:
             self.edge_,
             self.nc_,
             self.n_,
+            progress,
+            _get_threadpool(),
         )
         self.solved_ = True
 
@@ -262,6 +267,12 @@ class SASCAGraph:
         XOR_CST = 2
         LOOKUP = 3
         AND_CST = 4
+        ADD = 5
+        ADD_CST = 6
+        MUL = 7
+        MUL_CST = 8
+        property_map_binary = {"AND": (AND, AND_CST), "MUL": (MUL, MUL_CST)}
+        property_map_nary = {"XOR": (XOR, XOR_CST), "ADD": (ADD, ADD_CST)}
 
         # edge id
         self.edge_ = 0
@@ -316,18 +327,19 @@ class SASCAGraph:
                 self._share_edge(property, property["output"])
                 self._share_edge(property, property["inputs"][0])
 
-            elif property["property"] == "AND":
+            elif property["property"] in property_map_binary.keys():
+                op, op_cst = property_map_binary[property["property"]]
                 # If no public inputs
                 if all(x in self.var_ for x in property["inputs"]):
-                    property["func"] = AND
+                    property["func"] = op
                     self._share_edge(property, property["output"])
                     for i in property["inputs"]:
                         self._share_edge(property, i)
 
                 # if and with public input
                 elif len(property["inputs"]) == 2:
-                    # AND with one public
-                    property["func"] = AND_CST
+                    # OP with one public
+                    property["func"] = op_cst
 
                     self._share_edge(property, property["output"])
 
@@ -345,11 +357,12 @@ class SASCAGraph:
                     # merge public input in the property
                     property["values"] = self.publics_[property["inputs"][i[1]]]
 
-            elif property["property"] == "XOR":
+            elif property["property"] in property_map_nary.keys():
+                op, op_cst = property_map_nary[property["property"]]
                 # if no inputs are public
                 if all(x in self.var_ for x in property["inputs"]):
-                    # XOR with no public
-                    property["func"] = XOR
+                    # OP with no public
+                    property["func"] = op
 
                     # share edge with the output
                     self._share_edge(property, property["output"])
@@ -359,8 +372,8 @@ class SASCAGraph:
                         self._share_edge(property, i)
 
                 elif len(property["inputs"]) == 2:
-                    # XOR with one public
-                    property["func"] = XOR_CST
+                    # OP with one public
+                    property["func"] = op_cst
 
                     # which of both inputs is public
                     if property["inputs"][0] in self.var_:
@@ -378,8 +391,9 @@ class SASCAGraph:
                     property["values"] = self.publics_[property["inputs"][i[1]]]
 
                 else:
+                    key = property["property"]
                     raise ValueError(
-                        "XOR must have two operands when one operand is public."
+                        f"{key} must have two operands when one operand is public."
                     )
 
             else:
@@ -462,7 +476,7 @@ class SASCAGraphParser:
             self.errors.append("NC appears multiple times, can only appear once.")
         elif len(self.nc_decls) == 0:
             self.errors.append("NC not declared.")
-        elif self.nc_decls[0] not in range(1, 2 ** 16 + 1):
+        elif self.nc_decls[0] not in range(1, 2**16 + 1):
             self.errors.append("NC not in admissible range [1, 2^16].")
         else:
             self.nc = self.nc_decls[0]
@@ -508,6 +522,14 @@ class SASCAGraphParser:
         if "^" in prop:
             prop_kind = "XOR"
             inputs = prop.split("^")
+        elif "+" in prop:
+            prop_kind = "ADD"
+            inputs = prop.split("+")
+        elif "*" in prop:
+            prop_kind = "MUL"
+            inputs = prop.split("*")
+            if len(inputs) != 2:
+                raise SASCAGraphError("MUL not supported yet for more than 2 operands.")
         elif "&" in prop:
             prop_kind = "AND"
             inputs = prop.split("&")
